@@ -1,7 +1,7 @@
 import asyncio
 import logging
-import re
 import os
+import re
 from typing import Any, Dict, List, Optional, Tuple
 
 import aiohttp
@@ -11,6 +11,8 @@ from auth_proxy.auth_plugins.base import AuthPlugin
 
 logger = logging.getLogger(__name__)
 
+WS_GUID = "258EAFA5-E914-47DA-95CA-C5AB0DC85B11"  # WebSocket GUID for handshake
+
 
 class AuthProxy:
     """Authenticating reverse proxy."""
@@ -32,6 +34,11 @@ class AuthProxy:
                 # Register plugin paths
                 for plugin_path in plugin.get_plugin_paths():
                     path_key = plugin_path.path
+
+                    logger.debug(
+                        f"Registering plugin path: {path_key} for plugin {plugin_name}"
+                    )
+
                     if path_key in self.plugin_paths:
                         logger.warning(
                             f"Path '{path_key}' already registered by another plugin. Overriding."
@@ -61,6 +68,9 @@ class AuthProxy:
         self.backend_port = backend.get("port", 3000)  # Default port
         self.backend_socket = backend.get("socket")
 
+        # WebSocket configuration
+        self.enable_websockets = config.get("enable_websockets", True)
+
         # Path rules - precompile regex patterns
         self.paths = []
         for path_rule in config.get("paths", []):
@@ -204,6 +214,216 @@ class AuthProxy:
         }
         return messages.get(status_code, "Unknown")
 
+    async def _handle_websocket(
+        self,
+        reader: asyncio.StreamReader,
+        writer: asyncio.StreamWriter,
+        request_line: str,
+        headers: Dict[str, str],
+        authenticated: bool,
+        auth_headers: Dict[str, str],
+    ) -> None:
+        """
+        Handle a WebSocket upgrade request.
+
+        Args:
+            reader: Stream reader for the client connection
+            writer: Stream writer for the client connection
+            request_line: The HTTP request line
+            headers: The HTTP headers
+            authenticated: Whether the request is authenticated
+            auth_headers: Authentication headers to forward
+        """
+        if not authenticated:
+            # Send 401 Unauthorized response
+            writer.write(b"HTTP/1.1 401 Unauthorized\r\n")
+            writer.write(b"Content-Type: text/plain\r\n")
+            writer.write(b"Content-Length: 12\r\n")
+            writer.write(b"\r\n")
+            writer.write(b"Unauthorized")
+            await writer.drain()
+            return
+
+        # Parse the request line
+        method, path, protocol = request_line.split(" ")
+
+        # Verify required WebSocket headers
+        if "Upgrade" not in headers or headers["Upgrade"].lower() != "websocket":
+            logger.error(
+                "Invalid WebSocket upgrade request: missing or invalid Upgrade header"
+            )
+            writer.write(b"HTTP/1.1 400 Bad Request\r\n\r\n")
+            await writer.drain()
+            return
+
+        if (
+            "Connection" not in headers
+            or "upgrade" not in headers["Connection"].lower()
+        ):
+            logger.error(
+                "Invalid WebSocket upgrade request: missing or invalid Connection header"
+            )
+            writer.write(b"HTTP/1.1 400 Bad Request\r\n\r\n")
+            await writer.drain()
+            return
+
+        if "Sec-WebSocket-Key" not in headers:
+            logger.error(
+                "Invalid WebSocket upgrade request: missing Sec-WebSocket-Key header"
+            )
+            writer.write(b"HTTP/1.1 400 Bad Request\r\n\r\n")
+            await writer.drain()
+            return
+
+        # Connect to backend
+        try:
+            if self.backend_socket:
+                # Unix socket
+                backend_reader, backend_writer = await asyncio.open_unix_connection(
+                    path=self.backend_socket
+                )
+            else:
+                # TCP socket
+                backend_reader, backend_writer = await asyncio.open_connection(
+                    host=self.backend_host, port=self.backend_port
+                )
+
+            # Forward the WebSocket upgrade request to the backend
+            backend_request = f"{method} {path} {protocol}\r\n"
+            for key, value in headers.items():
+                backend_request += f"{key}: {value}\r\n"
+
+            # Add authentication headers
+            for key, value in auth_headers.items():
+                backend_request += f"{key}: {value}\r\n"
+
+            backend_request += "\r\n"
+            backend_writer.write(backend_request.encode())
+            await backend_writer.drain()
+
+            # Read the backend response
+            response_line = await backend_reader.readline()
+            if not response_line:
+                logger.error("Empty response from backend for WebSocket upgrade")
+                writer.write(b"HTTP/1.1 502 Bad Gateway\r\n\r\n")
+                await writer.drain()
+                return
+
+            # Parse the response status
+            response_line_str = response_line.decode("utf-8").strip()
+            protocol, status_code, *reason = response_line_str.split(" ", 2)
+
+            if status_code != "101":
+                # Backend did not accept the WebSocket upgrade, forward the response as-is
+                writer.write(response_line)
+
+                # Forward headers
+                while True:
+                    header_line = await backend_reader.readline()
+                    writer.write(header_line)
+                    if header_line == b"\r\n":
+                        break
+
+                # Forward body if any
+                while True:
+                    chunk = await backend_reader.read(4096)
+                    if not chunk:
+                        break
+                    writer.write(chunk)
+
+                await writer.drain()
+                return
+
+            # Backend accepted the upgrade, forward the 101 response to the client
+            writer.write(response_line)
+
+            # Forward headers from backend
+            backend_headers = {}
+            while True:
+                header_line = await backend_reader.readline()
+                if header_line == b"\r\n":
+                    writer.write(header_line)
+                    break
+
+                writer.write(header_line)
+
+                # Parse header for logging
+                try:
+                    header_line_str = header_line.decode("utf-8").strip()
+                    if ":" in header_line_str:
+                        key, value = header_line_str.split(":", 1)
+                        backend_headers[key.strip()] = value.strip()
+                except Exception:
+                    pass
+
+            await writer.drain()
+
+            # Now we have an established WebSocket connection
+            # Set up bidirectional forwarding
+            logger.info(f"WebSocket connection established for {path}")
+
+            # Create tasks for bidirectional forwarding
+            client_to_backend = asyncio.create_task(
+                self._forward_websocket_data(
+                    reader, backend_writer, "client -> backend"
+                )
+            )
+            backend_to_client = asyncio.create_task(
+                self._forward_websocket_data(
+                    backend_reader, writer, "backend -> client"
+                )
+            )
+
+            # Wait for either direction to complete (connection closed)
+            done, pending = await asyncio.wait(
+                [client_to_backend, backend_to_client],
+                return_when=asyncio.FIRST_COMPLETED,
+            )
+
+            # Cancel the remaining task
+            for task in pending:
+                task.cancel()
+
+            logger.info(f"WebSocket connection closed for {path}")
+
+        except Exception as e:
+            logger.error(f"Error handling WebSocket connection: {e}")
+            try:
+                writer.write(b"HTTP/1.1 502 Bad Gateway\r\n\r\n")
+                await writer.drain()
+            except:
+                pass
+
+    async def _forward_websocket_data(
+        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter, direction: str
+    ) -> None:
+        """
+        Forward WebSocket data between client and backend.
+
+        Args:
+            reader: Source stream reader
+            writer: Destination stream writer
+            direction: Direction label for logging
+        """
+        try:
+            while True:
+                data = await reader.read(8192)
+                if not data:
+                    break
+                writer.write(data)
+                await writer.drain()
+        except asyncio.CancelledError:
+            # Task was cancelled, this is normal when the other direction completes
+            pass
+        except Exception as e:
+            logger.error(f"Error forwarding WebSocket data ({direction}): {e}")
+        finally:
+            try:
+                writer.close()
+                await writer.wait_closed()
+            except:
+                pass
+
     async def handle_request(
         self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
     ) -> None:
@@ -241,6 +461,13 @@ class AuthProxy:
             content_length = int(headers.get("Content-Length", "0"))
             body = await reader.read(content_length) if content_length else b""
 
+            # Check if this is a WebSocket upgrade request
+            is_websocket = (
+                self.enable_websockets
+                and headers.get("Upgrade", "").lower() == "websocket"
+                and "upgrade" in headers.get("Connection", "").lower()
+            )
+
             # Check if this is a plugin-specific path
             plugin_name = None
             for plugin_path, path_info in self.plugin_paths.items():
@@ -261,6 +488,10 @@ class AuthProxy:
 
             # If this is a plugin-specific path, let the plugin handle it
             if plugin_name and plugin_name in self.auth_plugins:
+                logger.debug(
+                    f"Handling request with plugin '{plugin_name}' for path '{path}'"
+                )
+
                 plugin = self.auth_plugins[plugin_name]
                 result = plugin.handle_plugin_path(path, headers, body)
 
@@ -276,7 +507,7 @@ class AuthProxy:
                     writer.write(b"\r\n")
 
                     # Add Content-Length if not present
-                    if "Content-Length" not in response_headers:
+                    if response_body and "Content-Length" not in response_headers:
                         response_headers["Content-Length"] = str(len(response_body))
 
                     # Write response headers
@@ -290,6 +521,9 @@ class AuthProxy:
 
                     # Write response body
                     if response_body:
+                        if isinstance(response_body, str):
+                            response_body = response_body.encode("utf-8")
+
                         writer.write(response_body)
 
                     await writer.drain()
@@ -332,6 +566,13 @@ class AuthProxy:
                 await writer.drain()
                 return
 
+            # Handle WebSocket upgrade if needed
+            if is_websocket:
+                await self._handle_websocket(
+                    reader, writer, request_line, headers, authenticated, auth_headers
+                )
+                return
+
             # Update headers with authentication information
             if authenticated:
                 # Convert auth_headers to the same format as headers
@@ -356,7 +597,7 @@ class AuthProxy:
 
             # Forward the request to the backend
             try:
-                logger.debug(f"Forwarding request to backend")
+                logger.debug("Forwarding request to backend")
                 async with aiohttp.ClientSession() as session:
                     # Construct backend URL
                     if self.backend_socket:
